#include <iostream>
#include <fstream>
#include <stdexcept>
#include <algorithm>
#include <vector>
#include <cstring>
#include <cstdlib>
#include <set>
#include <array>

#include "SDL.h"
#include "SDL_vulkan.h"
#include "vulkan/vulkan.h"
#include "SDL_image.h"
#include "glm/glm.hpp"
#define VMA_IMPLEMENTATION
#include "vk_mem_alloc.h"

#define USE_VMA_FOR_PERSISTENT
#define USE_STAGING_IMAGE_BUFFER
const bool EnableValidationLayers = true;
const char* VertShaderPath = "test.vert.spv";
const char* FragShaderPath = "test.frag.spv";
const char* TexturePath = "texture.jpg";

class Application {
public:
    
    struct Vertex
    {
        glm::vec2 pos;
        glm::vec2 texCoord;
        glm::vec4 color;
    };
    const std::vector<Vertex> Vertices = {{{-1,-1}, {0,0}, {1,1,1,1}}, {{0,-1}, {1,0}, {1,1,1,1}}, {{0,0}, {1,1}, {1,1,1,1}}, {{-1,0}, {0,1}, {1,1,1,1}}};
    const std::vector<uint16_t> Indices = {0, 1, 2, 2, 3, 0};

    SDL_Window* m_window;
    VkInstance m_instance;
    VkDebugUtilsMessengerEXT m_debugCallback;
    VkSurfaceKHR m_surface;
    VkPhysicalDevice m_physicalDevice = VK_NULL_HANDLE;
    VkDevice m_device;
    uint32_t m_deviceGraphicsQueueFamilyIndex;
    uint32_t m_devicePresentQueueFamilyIndex;
    VkQueue m_graphicsQueue;
    VkQueue m_presentQueue;
    VmaAllocator m_allocator;

    VkSurfaceFormatKHR m_surfaceFormat;
    VkSwapchainKHR m_swapChain;
    std::vector<VkImage> m_swapChainImages;
    VkFormat m_swapChainImageFormat;
    VkExtent2D m_swapChainExtent;
    std::vector<VkImageView> m_swapChainImageViews;
    std::vector<VkFramebuffer> m_swapChainFramebuffers;

    VkRenderPass m_renderPass;
    VkPipelineLayout m_pipelineLayout;
    VkPipeline m_graphicsPipeline;

    VkCommandPool m_commandPool;
    VkCommandBuffer m_commandBuffer;
    VkDescriptorPool m_descriptorPool;
    VkDescriptorSet m_descriptorSet;
    VkSemaphore m_imageAvailableSemaphore;
    VkSemaphore m_renderFinishedSemaphore;
    VkFence m_inFlightFence;

    VkImage m_textureImage;
    VmaAllocation m_textureImageAllocation;
    VkImageView m_textureImageView;
    VkSampler m_textureSampler;
    
    VkBuffer m_vertexBuffer;
    VmaAllocation m_vertexBufferAllocation;
    VkBuffer m_indexBuffer;
    VmaAllocation m_indexBufferAllocation;
    
    VkBuffer m_testBuffer;
    VmaAllocation m_testBufferAllocation;
    VkDeviceMemory m_testBufferMemory;
    void* m_testBufferData;
    
    
    
    static VKAPI_ATTR VkBool32 VKAPI_CALL debugCallback(VkDebugUtilsMessageSeverityFlagBitsEXT messageSeverity, VkDebugUtilsMessageTypeFlagsEXT messageType, const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData, void* pUserData)
    {
        std::cout << "Validation Layer: " << pCallbackData->pMessage << std::endl;
        return VK_FALSE;
    }
    
    std::vector<char> readFile(const std::string& filename)
    {
        std::ifstream file(filename, std::ios::ate | std::ios::binary);
        size_t fileSize = (size_t) file.tellg();
        std::vector<char> buffer(fileSize);
        file.seekg(0);
        file.read(buffer.data(), fileSize);
        file.close();
        return buffer;
    }

    void mainLoop()
    {
        bool done = false;
        while (!done)
        {
            SDL_Event event;
            while (SDL_PollEvent(&event))
            {
                switch (event.type)
                {
                    case SDL_QUIT: done = true; break;
                    default: break;
                }
            }
            drawFrame();
        }
    }
    
    bool init()
    {
        SDL_Init(SDL_INIT_VIDEO);
        IMG_Init(IMG_INIT_PNG | IMG_INIT_JPG);
        m_window = SDL_CreateWindow("Vulkan", 0, 0, 800, 600, SDL_WINDOW_VULKAN | SDL_WINDOW_ALLOW_HIGHDPI | SDL_WINDOW_RESIZABLE);
        
        const std::vector<const char*> ValidationLayers = {"VK_LAYER_LUNARG_standard_validation"};
        const std::vector<const char*> DeviceExtensions = {VK_KHR_SWAPCHAIN_EXTENSION_NAME};
        
        //create instance
        VkApplicationInfo appInfo = {VK_STRUCTURE_TYPE_APPLICATION_INFO, nullptr, "", 1, "", 1, VK_API_VERSION_1_0};
        VkInstanceCreateInfo instanceCreateInfo = {};
        instanceCreateInfo.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
        instanceCreateInfo.pApplicationInfo = &appInfo;
        unsigned int extensionCount = 0;
        SDL_Vulkan_GetInstanceExtensions(m_window, &extensionCount, nullptr);
        std::vector<const char*> extensions(extensionCount);
        SDL_Vulkan_GetInstanceExtensions(m_window, &extensionCount, extensions.data());
        if (EnableValidationLayers)
            extensions.push_back(VK_EXT_DEBUG_UTILS_EXTENSION_NAME);
        instanceCreateInfo.enabledExtensionCount = static_cast<uint32_t>(extensions.size());
        instanceCreateInfo.ppEnabledExtensionNames = extensions.data();
        if (EnableValidationLayers)
        {
            uint32_t layerCount;
            vkEnumerateInstanceLayerProperties(&layerCount, nullptr);
            std::vector<VkLayerProperties> availableLayers(layerCount);
            vkEnumerateInstanceLayerProperties(&layerCount, availableLayers.data());
            instanceCreateInfo.enabledLayerCount = static_cast<uint32_t>(ValidationLayers.size());
            instanceCreateInfo.ppEnabledLayerNames = ValidationLayers.data();
        }
        else
            instanceCreateInfo.enabledLayerCount = 0;
        vkCreateInstance(&instanceCreateInfo, nullptr, &m_instance);
        
        if (EnableValidationLayers)
        {
            VkDebugUtilsMessengerCreateInfoEXT createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT;
            createInfo.messageSeverity = VK_DEBUG_UTILS_MESSAGE_SEVERITY_VERBOSE_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT;
            createInfo.messageType = VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT;
            createInfo.pfnUserCallback = debugCallback;
            auto func = (PFN_vkCreateDebugUtilsMessengerEXT)vkGetInstanceProcAddr(m_instance, "vkCreateDebugUtilsMessengerEXT");
            func(m_instance, &createInfo, nullptr, &m_debugCallback);
        }
        
        SDL_Vulkan_CreateSurface(m_window, m_instance, &m_surface);
        
        //create device
        uint32_t deviceCount = 0;
        vkEnumeratePhysicalDevices(m_instance, &deviceCount, nullptr);
        std::vector<VkPhysicalDevice> devices(deviceCount);
        vkEnumeratePhysicalDevices(m_instance, &deviceCount, devices.data());
        for (const auto& device : devices)
        {
            int graphicsQueueFamilyIndex = -1;
            int presentQueueFamilyIndex = -1;
            auto hasQueueFamilies = [&graphicsQueueFamilyIndex,&presentQueueFamilyIndex]{ return graphicsQueueFamilyIndex >= 0 && presentQueueFamilyIndex >= 0; };
            uint32_t queueFamilyCount = 0;
            vkGetPhysicalDeviceQueueFamilyProperties(device, &queueFamilyCount, nullptr);
            std::vector<VkQueueFamilyProperties> queueFamilies(queueFamilyCount);
            vkGetPhysicalDeviceQueueFamilyProperties(device, &queueFamilyCount, queueFamilies.data());
            for (int i=0;i<queueFamilies.size();++i)
            {
                if (queueFamilies[i].queueCount > 0)
                {
                    if (queueFamilies[i].queueFlags & VK_QUEUE_GRAPHICS_BIT)
                        graphicsQueueFamilyIndex = i;
                    VkBool32 presentSupport = false;
                    vkGetPhysicalDeviceSurfaceSupportKHR(device, i, m_surface, &presentSupport);
                    if (presentSupport)
                        presentQueueFamilyIndex = i;
                }
                if (hasQueueFamilies())
                    break;
            }
            if (!hasQueueFamilies())
                continue;
            uint32_t extensionCount;
            vkEnumerateDeviceExtensionProperties(device, nullptr, &extensionCount, nullptr);
            std::vector<VkExtensionProperties> availableExtensions(extensionCount);
            vkEnumerateDeviceExtensionProperties(device, nullptr, &extensionCount, availableExtensions.data());
            bool allDeviceExtensionsAvailable = true;
            for (const auto& extension : DeviceExtensions)
                allDeviceExtensionsAvailable &= std::find_if(availableExtensions.begin(), availableExtensions.end(), [extension](const VkExtensionProperties& p){ return std::strcmp(extension, p.extensionName) == 0; }) != availableExtensions.end();
            if (!allDeviceExtensionsAvailable)
                continue;
            uint32_t formatCount, presentModeCount;
            vkGetPhysicalDeviceSurfaceFormatsKHR(device, m_surface, &formatCount, nullptr);
            vkGetPhysicalDeviceSurfacePresentModesKHR(device, m_surface, &presentModeCount, nullptr);
            if (formatCount == 0 || presentModeCount == 0)
                continue;
            std::vector<VkSurfaceFormatKHR> surfaceFormats(formatCount);
            vkGetPhysicalDeviceSurfaceFormatsKHR(device, m_surface, &formatCount, surfaceFormats.data());
            std::vector<VkPresentModeKHR> surfacePresentModes(presentModeCount);
            vkGetPhysicalDeviceSurfacePresentModesKHR(device, m_surface, &presentModeCount, surfacePresentModes.data());
            VkPhysicalDeviceFeatures supportedFeatures;
            vkGetPhysicalDeviceFeatures(device, &supportedFeatures);
            if (!supportedFeatures.samplerAnisotropy)
                continue;
            
            m_physicalDevice = device;
            m_deviceGraphicsQueueFamilyIndex = graphicsQueueFamilyIndex;
            m_devicePresentQueueFamilyIndex = presentQueueFamilyIndex;
            m_surfaceFormat = surfaceFormats[0];
            if (surfaceFormats.size() == 1 && surfaceFormats[0].format == VK_FORMAT_UNDEFINED)
                m_surfaceFormat = {VK_FORMAT_B8G8R8A8_UNORM, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR};
            else
            {
                for (const auto& sf : surfaceFormats)
                {
                    if (sf.format == VK_FORMAT_B8G8R8A8_UNORM && sf.colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR)
                    {
                        m_surfaceFormat = sf;
                        break;
                    }
                }
            }
            break;
        }
        std::vector<VkDeviceQueueCreateInfo> queueCreateInfos;
        std::set<uint32_t> uniqueQueueFamilies = {m_deviceGraphicsQueueFamilyIndex, m_devicePresentQueueFamilyIndex};
        float queuePriority = 1.0f;
        for (auto queueFamily : uniqueQueueFamilies)
            queueCreateInfos.push_back({VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO, nullptr, 0, queueFamily, 1, &queuePriority});
        VkDeviceCreateInfo deviceCreateInfo = {VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO, nullptr, 0, (uint32_t)queueCreateInfos.size(), queueCreateInfos.data(), 0, nullptr, (uint32_t)DeviceExtensions.size(), DeviceExtensions.data(), nullptr};
        if (EnableValidationLayers)
        {
            deviceCreateInfo.enabledLayerCount = static_cast<uint32_t>(ValidationLayers.size());
            deviceCreateInfo.ppEnabledLayerNames = ValidationLayers.data();
        }
        vkCreateDevice(m_physicalDevice, &deviceCreateInfo, nullptr, &m_device);
        vkGetDeviceQueue(m_device, m_deviceGraphicsQueueFamilyIndex, 0, &m_graphicsQueue);
        vkGetDeviceQueue(m_device, m_devicePresentQueueFamilyIndex, 0, &m_presentQueue);
        
        VmaAllocatorCreateInfo allocatorInfo = {};
        allocatorInfo.physicalDevice = m_physicalDevice;
        allocatorInfo.device = m_device;
        vmaCreateAllocator(&allocatorInfo, &m_allocator);
        
        VkShaderModule vertShaderModule, fragShaderModule;
        std::vector<char> vertShaderData = readFile(VertShaderPath);
        std::vector<char> fragShaderData = readFile(FragShaderPath);
        VkShaderModuleCreateInfo shaderModuleCreateInfo = {VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO, nullptr, 0, vertShaderData.size(), reinterpret_cast<const uint32_t*>(vertShaderData.data())};
        vkCreateShaderModule(m_device, &shaderModuleCreateInfo, nullptr, &vertShaderModule);
        shaderModuleCreateInfo = {VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO, nullptr, 0, fragShaderData.size(), reinterpret_cast<const uint32_t*>(fragShaderData.data())};
        vkCreateShaderModule(m_device, &shaderModuleCreateInfo, nullptr, &fragShaderModule);
        std::vector<VkPipelineShaderStageCreateInfo> shaderStages;
        shaderStages.push_back({VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO, nullptr, 0, VK_SHADER_STAGE_VERTEX_BIT, vertShaderModule, "main", nullptr});
        shaderStages.push_back({VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO, nullptr, 0, VK_SHADER_STAGE_FRAGMENT_BIT, fragShaderModule, "main", nullptr});
        
        VkDescriptorSetLayout descriptorSetLayout;
        VkDescriptorSetLayoutBinding binding = {0, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 1, VK_SHADER_STAGE_FRAGMENT_BIT, nullptr};
        VkDescriptorSetLayoutCreateInfo layoutInfo = {VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO, nullptr, 0, 1, &binding};
        vkCreateDescriptorSetLayout(m_device, &layoutInfo, nullptr, &descriptorSetLayout);
        VkPipelineLayoutCreateInfo pipelineLayoutInfo = {VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO, nullptr, 0, 1, &descriptorSetLayout, 0, nullptr};
        vkCreatePipelineLayout(m_device, &pipelineLayoutInfo, nullptr, &m_pipelineLayout);
        VkDescriptorPoolSize descriptorPoolSize = {VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 1};
        VkDescriptorPoolCreateInfo descriptorPoolInfo = {VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO, nullptr, 0, 1, 1, &descriptorPoolSize};
        vkCreateDescriptorPool(m_device, &descriptorPoolInfo, nullptr, &m_descriptorPool);
        VkDescriptorSetAllocateInfo descriptorAllocInfo = {VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO, nullptr, m_descriptorPool, 1, &descriptorSetLayout};
        vkAllocateDescriptorSets(m_device, &descriptorAllocInfo, &m_descriptorSet);
        VkCommandPoolCreateInfo commandPoolInfo = {VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO, nullptr, VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT | VK_COMMAND_POOL_CREATE_TRANSIENT_BIT, m_deviceGraphicsQueueFamilyIndex};
        vkCreateCommandPool(m_device, &commandPoolInfo, nullptr, &m_commandPool);
        VkCommandBufferAllocateInfo allocInfo = {VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO, nullptr, m_commandPool, VK_COMMAND_BUFFER_LEVEL_PRIMARY, 1};
        vkAllocateCommandBuffers(m_device, &allocInfo, &m_commandBuffer);
        VkSemaphoreCreateInfo semaphoreInfo = {VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO, nullptr, 0};
        vkCreateSemaphore(m_device, &semaphoreInfo, nullptr, &m_imageAvailableSemaphore);
        vkCreateSemaphore(m_device, &semaphoreInfo, nullptr, &m_renderFinishedSemaphore);
        VkFenceCreateInfo fenceInfo = {VK_STRUCTURE_TYPE_FENCE_CREATE_INFO, nullptr, VK_FENCE_CREATE_SIGNALED_BIT};
        vkCreateFence(m_device, &fenceInfo, nullptr, &m_inFlightFence);
        
        //create swapChain
        VkSurfaceCapabilitiesKHR surfaceCapabilities;
        vkGetPhysicalDeviceSurfaceCapabilitiesKHR(m_physicalDevice, m_surface, &surfaceCapabilities);
        VkExtent2D extent;
        if (surfaceCapabilities.currentExtent.width != std::numeric_limits<uint32_t>::max())
            extent = surfaceCapabilities.currentExtent;
        else
        {
            int width, height;
            SDL_Vulkan_GetDrawableSize(m_window, &width, &height);
            extent = {std::clamp(static_cast<uint32_t>(width), surfaceCapabilities.minImageExtent.width, surfaceCapabilities.maxImageExtent.width), std::clamp(static_cast<uint32_t>(height), surfaceCapabilities.minImageExtent.height, surfaceCapabilities.maxImageExtent.height)};
        }
        uint32_t imageCount = surfaceCapabilities.minImageCount + 1;
        if (surfaceCapabilities.maxImageCount > 0 && imageCount > surfaceCapabilities.maxImageCount)
            imageCount = surfaceCapabilities.maxImageCount;
        VkSwapchainCreateInfoKHR createInfo = {VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR, nullptr, 0, m_surface, imageCount, m_surfaceFormat.format, m_surfaceFormat.colorSpace, extent, 1, VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT, VK_SHARING_MODE_EXCLUSIVE, 0, nullptr, surfaceCapabilities.currentTransform, VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR, VK_PRESENT_MODE_FIFO_KHR, VK_TRUE, VK_NULL_HANDLE};
        std::vector<uint32_t> queueFamiliyIndices = {m_deviceGraphicsQueueFamilyIndex, m_devicePresentQueueFamilyIndex};
        if (m_deviceGraphicsQueueFamilyIndex != m_devicePresentQueueFamilyIndex)
        {
            createInfo.imageSharingMode = VK_SHARING_MODE_CONCURRENT;
            createInfo.queueFamilyIndexCount = static_cast<uint32_t>(queueFamiliyIndices.size());
            createInfo.pQueueFamilyIndices = queueFamiliyIndices.data();
        }
        vkCreateSwapchainKHR(m_device, &createInfo, nullptr, &m_swapChain);
        vkGetSwapchainImagesKHR(m_device, m_swapChain, &imageCount, nullptr);
        m_swapChainImages.resize(imageCount);
        vkGetSwapchainImagesKHR(m_device, m_swapChain, &imageCount, m_swapChainImages.data());
        m_swapChainExtent = extent;
        m_swapChainImageFormat = m_surfaceFormat.format;
        
        //renderpass
        VkAttachmentDescription colorAttachment = {0, m_swapChainImageFormat, VK_SAMPLE_COUNT_1_BIT, VK_ATTACHMENT_LOAD_OP_CLEAR, VK_ATTACHMENT_STORE_OP_STORE, VK_ATTACHMENT_LOAD_OP_DONT_CARE, VK_ATTACHMENT_STORE_OP_DONT_CARE, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_PRESENT_SRC_KHR};
        VkAttachmentReference colorAttachmentRef = {0, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL};
        VkSubpassDescription subpass = {0, VK_PIPELINE_BIND_POINT_GRAPHICS, 0, nullptr, 1, &colorAttachmentRef, nullptr, nullptr, 0, nullptr};
        VkSubpassDependency dependency = {VK_SUBPASS_EXTERNAL, 0, VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT, VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT, 0, VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT, 0};
        VkRenderPassCreateInfo renderPassCreateInfo = {VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO, nullptr, 0, 1, &colorAttachment, 1, &subpass, 1, &dependency};
        vkCreateRenderPass(m_device, &renderPassCreateInfo, nullptr, &m_renderPass);
        
        //graphics pipeline
        VkVertexInputBindingDescription bindingDescription = {0, sizeof(Vertex), VK_VERTEX_INPUT_RATE_VERTEX};
        std::vector<VkVertexInputAttributeDescription> attributeDescriptions = {{0,0,VK_FORMAT_R32G32_SFLOAT,offsetof(Vertex, pos)}, {1,0,VK_FORMAT_R32G32_SFLOAT,offsetof(Vertex, texCoord)}, {2,0,VK_FORMAT_R32G32B32A32_SFLOAT,offsetof(Vertex, color)}};
        VkPipelineVertexInputStateCreateInfo vertexInputInfo = {VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO, nullptr, 0, 1, &bindingDescription, (uint32_t)attributeDescriptions.size(), attributeDescriptions.data()};
        VkPipelineInputAssemblyStateCreateInfo inputAssembly = {VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO, nullptr, 0, VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST, VK_FALSE};
        VkViewport viewport = {0,0,(float)m_swapChainExtent.width,(float)m_swapChainExtent.height,0,1};
        VkRect2D scissor = {{0,0}, m_swapChainExtent};
        VkPipelineViewportStateCreateInfo viewportState = {VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO, nullptr, 0, 1, &viewport, 1, &scissor};
        VkPipelineRasterizationStateCreateInfo rasterizer = {VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO, nullptr, 0, VK_FALSE, VK_FALSE, VK_POLYGON_MODE_FILL, VK_CULL_MODE_NONE, VK_FRONT_FACE_COUNTER_CLOCKWISE, VK_FALSE, 0, 0, 0, 1};
        VkPipelineMultisampleStateCreateInfo multisampling = {VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO, nullptr, 0, VK_SAMPLE_COUNT_1_BIT, VK_FALSE, 0, nullptr, 0, 0};
        VkPipelineColorBlendAttachmentState colorBlendAttachment = {VK_FALSE, VK_BLEND_FACTOR_ZERO, VK_BLEND_FACTOR_ZERO, VK_BLEND_OP_ADD, VK_BLEND_FACTOR_ZERO, VK_BLEND_FACTOR_ZERO, VK_BLEND_OP_ADD, VK_COLOR_COMPONENT_R_BIT|VK_COLOR_COMPONENT_G_BIT|VK_COLOR_COMPONENT_B_BIT|VK_COLOR_COMPONENT_A_BIT};
        VkPipelineColorBlendStateCreateInfo colorBlending = {VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO, nullptr, 0, VK_FALSE, VK_LOGIC_OP_COPY, 1, &colorBlendAttachment, {0,0,0,0}};
        VkGraphicsPipelineCreateInfo pipelineInfo = {VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO, nullptr, 0, (uint32_t)shaderStages.size(), shaderStages.data(), &vertexInputInfo, &inputAssembly, nullptr, &viewportState, &rasterizer, &multisampling, nullptr, &colorBlending, nullptr, m_pipelineLayout, m_renderPass, 0, VK_NULL_HANDLE, 0};
        vkCreateGraphicsPipelines(m_device, VK_NULL_HANDLE, 1, &pipelineInfo, nullptr, &m_graphicsPipeline);
        
        //create swapChain image views and framebuffers
        m_swapChainImageViews.resize(m_swapChainImages.size());
        for (uint32_t i = 0; i < m_swapChainImages.size(); i++)
            m_swapChainImageViews[i] = createImageView(m_swapChainImages[i], m_swapChainImageFormat, VK_IMAGE_ASPECT_COLOR_BIT);
        m_swapChainFramebuffers.resize(m_swapChainImageViews.size());
        for (size_t i = 0; i < m_swapChainImageViews.size(); i++)
        {
            VkFramebufferCreateInfo framebufferInfo = {VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO, nullptr, 0, m_renderPass, 1, &m_swapChainImageViews[i], m_swapChainExtent.width, m_swapChainExtent.height, 1};
            vkCreateFramebuffer(m_device, &framebufferInfo, nullptr, &m_swapChainFramebuffers[i]);
        }
        
        return true;
    }
    
    void beginSingleTimeCommands()
    {
        VkCommandBufferBeginInfo beginInfo = {};
        beginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
        beginInfo.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
        if (vkBeginCommandBuffer(m_commandBuffer, &beginInfo) != VK_SUCCESS)
            throw std::runtime_error("beginSingleTimeCommands");
    }
    
    void endSingleTimeCommands()
    {
        vkEndCommandBuffer(m_commandBuffer);
        VkSubmitInfo submitInfo = {VK_STRUCTURE_TYPE_SUBMIT_INFO, nullptr, 0, nullptr, nullptr, 1, &m_commandBuffer, 0, nullptr};
        vkQueueSubmit(m_graphicsQueue, 1, &submitInfo, VK_NULL_HANDLE);
        vkQueueWaitIdle(m_graphicsQueue);
    }
    
    void createImage(uint32_t width, uint32_t height, VkSampleCountFlagBits numSamples, VkFormat format, VkImageTiling tiling, VkImageLayout initialLayout, VkImageUsageFlags usage, VkMemoryPropertyFlags properties, VkImage& image, VmaAllocation& allocation)
    {
        VkImageCreateInfo imageInfo = {VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO, nullptr, 0, VK_IMAGE_TYPE_2D, format, {width, height, 1}, 1, 1, numSamples, tiling, usage, VK_SHARING_MODE_EXCLUSIVE, 0, nullptr, initialLayout};
        VmaAllocationCreateInfo allocInfo = {};
        allocInfo.requiredFlags = properties;
        vmaCreateImage(m_allocator, &imageInfo, &allocInfo, &image, &allocation, nullptr);
    }
    
    VkImageView createImageView(VkImage image, VkFormat format, VkImageAspectFlags aspectFlags)
    {
        VkImageView imageView;
        VkImageViewCreateInfo viewInfo = {VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO, nullptr, 0, image, VK_IMAGE_VIEW_TYPE_2D, format, {}, {aspectFlags, 0, 1, 0, 1}};
        vkCreateImageView(m_device, &viewInfo, nullptr, &imageView);
        return imageView;
    }
    
    void createBuffer(VkDeviceSize size, VkBufferUsageFlags usage, VkMemoryPropertyFlags properties, VmaAllocationCreateFlags createFlags, VkBuffer& buffer, VmaAllocation& allocation)
    {
        VkBufferCreateInfo bufferInfo = {VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO, nullptr, 0, size, usage, VK_SHARING_MODE_EXCLUSIVE, 0, nullptr};
        VmaAllocationCreateInfo allocInfo = {};
        allocInfo.requiredFlags = properties;
        allocInfo.flags = createFlags;
        vmaCreateBuffer(m_allocator, &bufferInfo, &allocInfo, &buffer, &allocation, nullptr);
    }
    
    void createBuffer(VkDeviceSize size, VkBufferUsageFlags usage, VkMemoryPropertyFlags properties, VkBuffer& buffer, VkDeviceMemory& bufferMemory)
    {
        VkBufferCreateInfo bufferInfo = {VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO, nullptr, 0, size, usage, VK_SHARING_MODE_EXCLUSIVE, 0, nullptr};
        vkCreateBuffer(m_device, &bufferInfo, nullptr, &buffer);
        VkMemoryRequirements memRequirements;
        vkGetBufferMemoryRequirements(m_device, buffer, &memRequirements);
        VkMemoryAllocateInfo allocInfo = {};
        allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
        allocInfo.allocationSize = memRequirements.size;
        allocInfo.memoryTypeIndex = findMemoryType(memRequirements.memoryTypeBits, properties);
        vkAllocateMemory(m_device, &allocInfo, nullptr, &bufferMemory);
        vkBindBufferMemory(m_device, buffer, bufferMemory, 0);
    }
    
    uint32_t findMemoryType(uint32_t typeFilter, VkMemoryPropertyFlags properties) {
        VkPhysicalDeviceMemoryProperties memProperties;
        vkGetPhysicalDeviceMemoryProperties(m_physicalDevice, &memProperties);
        for (uint32_t i = 0; i < memProperties.memoryTypeCount; i++)
            if ((typeFilter & (1 << i)) && (memProperties.memoryTypes[i].propertyFlags & properties) == properties)
                return i;
        return 0;
    }
    
    void transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout)
    {
        auto setFlags = [](VkImageLayout layout, VkAccessFlags& accessMask, VkPipelineStageFlags& stage)
        {
            switch (layout)
            {
                case VK_IMAGE_LAYOUT_UNDEFINED: accessMask = 0; stage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT; break;
                case VK_IMAGE_LAYOUT_PREINITIALIZED: accessMask = VK_ACCESS_HOST_WRITE_BIT; stage = VK_PIPELINE_STAGE_HOST_BIT; break;
                case VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL: accessMask = VK_ACCESS_TRANSFER_READ_BIT; stage = VK_PIPELINE_STAGE_TRANSFER_BIT; break;
                case VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL: accessMask = VK_ACCESS_TRANSFER_WRITE_BIT; stage = VK_PIPELINE_STAGE_TRANSFER_BIT; break;
                case VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL: accessMask = VK_ACCESS_SHADER_READ_BIT; stage = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT; break;
                case VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL: accessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT; stage = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT; break;
                case VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL: accessMask = VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_READ_BIT | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT; stage = VK_PIPELINE_STAGE_EARLY_FRAGMENT_TESTS_BIT; break;
                default: throw std::runtime_error("transitionImageLayout"); break;
            }
        };
        VkImageMemoryBarrier barrier = {VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER, nullptr, 0, 0, oldLayout, newLayout, VK_QUEUE_FAMILY_IGNORED, VK_QUEUE_FAMILY_IGNORED, image, {VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1}};
        VkPipelineStageFlags sourceStage, destinationStage;
        setFlags(oldLayout, barrier.srcAccessMask, sourceStage);
        setFlags(newLayout, barrier.dstAccessMask, destinationStage);
        beginSingleTimeCommands();
        vkCmdPipelineBarrier(m_commandBuffer, sourceStage, destinationStage, 0, 0, nullptr, 0, nullptr, 1, &barrier);
        endSingleTimeCommands();
    }
    
    void copyBuffer(VkBuffer srcBuffer, VkBuffer dstBuffer, VkDeviceSize size)
    {
        beginSingleTimeCommands();
        VkBufferCopy copyRegion = {{}, {}, size};
        vkCmdCopyBuffer(m_commandBuffer, srcBuffer, dstBuffer, 1, &copyRegion);
        endSingleTimeCommands();
    }
    
    void copyImage(VkImage srcImage, VkImage dstImage, uint32_t width, uint32_t height)
    {
        beginSingleTimeCommands();
        VkImageSubresourceLayers subresource = {VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1};
        VkImageCopy region = {subresource, {0, 0, 0}, subresource, {0, 0, 0}, {width, height, 1}};
        vkCmdCopyImage(m_commandBuffer, srcImage, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, dstImage, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &region);
        endSingleTimeCommands();
    }
    
    void copyBufferToImage(VkBuffer buffer, VkImage image, uint32_t width, uint32_t height)
    {
        beginSingleTimeCommands();
        VkBufferImageCopy region = {0, 0, 0, {VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1}, {0,0,0}, {width, height, 1}};
        vkCmdCopyBufferToImage(m_commandBuffer, buffer, image, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &region);
        endSingleTimeCommands();
    }
    
    void createTextureImage()
    {
        SDL_Surface* surface = IMG_Load(TexturePath);
        if (surface->format->format != SDL_PIXELFORMAT_ABGR8888)
        {
            SDL_Surface* convertedSurface = SDL_ConvertSurfaceFormat(surface, SDL_PIXELFORMAT_ABGR8888, 0);
            SDL_FreeSurface(surface);
            surface = convertedSurface;
        }
        SDL_LockSurface(surface);
        createImage(surface->w, surface->h, VK_SAMPLE_COUNT_1_BIT, VK_FORMAT_R8G8B8A8_UNORM, VK_IMAGE_TILING_OPTIMAL, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, m_textureImage, m_textureImageAllocation);
        transitionImageLayout(m_textureImage, VK_FORMAT_R8G8B8A8_UNORM, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);
        void* imageData;
        
#ifdef USE_STAGING_IMAGE_BUFFER
        VkImage stagingImage;
        VmaAllocation stagingImageMemory;
        createImage(surface->w, surface->h, VK_SAMPLE_COUNT_1_BIT, VK_FORMAT_R8G8B8A8_UNORM, VK_IMAGE_TILING_LINEAR, VK_IMAGE_LAYOUT_PREINITIALIZED, VK_IMAGE_USAGE_TRANSFER_SRC_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, stagingImage, stagingImageMemory);
        VkImageSubresource subresource = {};
        subresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        subresource.mipLevel = 0;
        subresource.arrayLayer = 0;
        VkSubresourceLayout stagingImageLayout;
        vkGetImageSubresourceLayout(m_device, stagingImage, &subresource, &stagingImageLayout);
        vmaMapMemory(m_allocator, stagingImageMemory, &imageData);
        if (stagingImageLayout.rowPitch == surface->w * 4)
            memcpy(imageData, surface->pixels, static_cast<size_t>(surface->w * surface->h * 4));
        else
        {
            uint8_t* dataBytes = reinterpret_cast<uint8_t*>(imageData);
            uint8_t* pixelBytes = reinterpret_cast<uint8_t*>(surface->pixels);
            for (int y = 0; y < surface->h; y++)
                memcpy(&dataBytes[y * stagingImageLayout.rowPitch], &pixelBytes[y * surface->w * 4], surface->w * 4);
        }
        vmaUnmapMemory(m_allocator, stagingImageMemory);
        transitionImageLayout(stagingImage, VK_FORMAT_R8G8B8A8_UNORM, VK_IMAGE_LAYOUT_PREINITIALIZED, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL);
        copyImage(stagingImage, m_textureImage, surface->w, surface->h);
        vmaDestroyImage(m_allocator, stagingImage, stagingImageMemory);
#else
        VkDeviceSize imageSize = surface->w * surface->h * 4;
        VkBuffer imageStagingBuffer;
        VmaAllocation imageStagingBufferAllocation;
        createBuffer(imageSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, 0, imageStagingBuffer, imageStagingBufferAllocation);
        vmaMapMemory(m_allocator, imageStagingBufferAllocation, &imageData);
        memcpy(imageData, surface->pixels, static_cast<size_t>(imageSize));
        vmaUnmapMemory(m_allocator, imageStagingBufferAllocation);
        copyBufferToImage(imageStagingBuffer, m_textureImage, surface->w, surface->h);
        vmaDestroyBuffer(m_allocator, imageStagingBuffer, imageStagingBufferAllocation);
#endif
        SDL_UnlockSurface(surface);
        SDL_FreeSurface(surface);
        transitionImageLayout(m_textureImage, VK_FORMAT_R8G8B8A8_UNORM, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL);
        
        
        //create sampler and update descriptor set
        m_textureImageView = createImageView(m_textureImage, VK_FORMAT_R8G8B8A8_UNORM, VK_IMAGE_ASPECT_COLOR_BIT);
        VkSamplerCreateInfo samplerInfo = {VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO, nullptr, 0, VK_FILTER_LINEAR, VK_FILTER_LINEAR, VK_SAMPLER_MIPMAP_MODE_LINEAR, VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, 0, VK_FALSE, 0, VK_FALSE, VK_COMPARE_OP_ALWAYS, 0, 1, VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK, 0};
        vkCreateSampler(m_device, &samplerInfo, nullptr, &m_textureSampler);
        VkDescriptorImageInfo imageInfo = {m_textureSampler, m_textureImageView, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL};
        VkWriteDescriptorSet descriptorWrite = {VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET, nullptr, m_descriptorSet, 0, 0, 1, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, &imageInfo, nullptr, nullptr};
        vkUpdateDescriptorSets(m_device, 1, &descriptorWrite, 0, nullptr);
        
        //create vertex and index buffers
        VkDeviceSize bufferSize = sizeof(Vertices[0]) * Vertices.size();
        VkBuffer stagingBuffer;
        VmaAllocation stagingBufferAllocation;
        createBuffer(bufferSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, 0, stagingBuffer, stagingBufferAllocation);
        void* data;
        vmaMapMemory(m_allocator, stagingBufferAllocation, &data);
        memcpy(data, Vertices.data(), (size_t)bufferSize);
        vmaUnmapMemory(m_allocator, stagingBufferAllocation);
        createBuffer(bufferSize, VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_VERTEX_BUFFER_BIT, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, 0, m_vertexBuffer, m_vertexBufferAllocation);
        copyBuffer(stagingBuffer, m_vertexBuffer, bufferSize);
        vmaDestroyBuffer(m_allocator, stagingBuffer, stagingBufferAllocation);
        bufferSize = sizeof(Indices[0]) * Indices.size();
        createBuffer(bufferSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, 0, stagingBuffer, stagingBufferAllocation);
        vmaMapMemory(m_allocator, stagingBufferAllocation, &data);
        memcpy(data, Indices.data(), (size_t) bufferSize);
        vmaUnmapMemory(m_allocator, stagingBufferAllocation);
        createBuffer(bufferSize, VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_INDEX_BUFFER_BIT, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, 0, m_indexBuffer, m_indexBufferAllocation);
        copyBuffer(stagingBuffer, m_indexBuffer, bufferSize);
        vmaDestroyBuffer(m_allocator, stagingBuffer, stagingBufferAllocation);
    }

    void drawFrame() {
        vkWaitForFences(m_device, 1, &m_inFlightFence, VK_TRUE, std::numeric_limits<uint64_t>::max());
        vkResetFences(m_device, 1, &m_inFlightFence);
        uint32_t imageIndex;
        vkAcquireNextImageKHR(m_device, m_swapChain, std::numeric_limits<uint64_t>::max(), m_imageAvailableSemaphore, VK_NULL_HANDLE, &imageIndex);
        
        VkCommandBufferBeginInfo beginInfo = {VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO, nullptr, VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT, nullptr};
        vkBeginCommandBuffer(m_commandBuffer, &beginInfo);
        VkClearValue clearValue = {0.0f, 0.0f, 0.0f, 1.0f};
        VkRenderPassBeginInfo renderPassInfo = {VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO, nullptr, m_renderPass, m_swapChainFramebuffers[imageIndex], {{0,0}, m_swapChainExtent}, 1, &clearValue};
        vkCmdBeginRenderPass(m_commandBuffer, &renderPassInfo, VK_SUBPASS_CONTENTS_INLINE);
        vkCmdBindPipeline(m_commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, m_graphicsPipeline);
        VkDeviceSize offsets[] = {0};
        vkCmdBindVertexBuffers(m_commandBuffer, 0, 1, &m_vertexBuffer, offsets);
        vkCmdBindIndexBuffer(m_commandBuffer, m_indexBuffer, 0, VK_INDEX_TYPE_UINT16);
        vkCmdBindDescriptorSets(m_commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, m_pipelineLayout, 0, 1, &m_descriptorSet, 0, nullptr);
        vkCmdDrawIndexed(m_commandBuffer, static_cast<uint32_t>(Indices.size()), 1, 0, 0, 0);
        vkCmdEndRenderPass(m_commandBuffer);
        vkEndCommandBuffer(m_commandBuffer);

        VkPipelineStageFlags waitStages[] = {VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT};
        VkSubmitInfo submitInfo = {VK_STRUCTURE_TYPE_SUBMIT_INFO, nullptr, 1, &m_imageAvailableSemaphore, waitStages, 1, &m_commandBuffer, 1, &m_renderFinishedSemaphore};
        vkQueueSubmit(m_graphicsQueue, 1, &submitInfo, m_inFlightFence);

        VkPresentInfoKHR presentInfo = {VK_STRUCTURE_TYPE_PRESENT_INFO_KHR, nullptr, 1, &m_renderFinishedSemaphore, 1, &m_swapChain, &imageIndex, nullptr};
        vkQueuePresentKHR(m_presentQueue, &presentInfo);
    }
    
    void mapPersistentMemory()
    {
#ifdef USE_VMA_FOR_PERSISTENT
        createBuffer(1, VK_BUFFER_USAGE_VERTEX_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, 0, m_testBuffer, m_testBufferAllocation);
        vmaMapMemory(m_allocator, m_testBufferAllocation, &m_testBufferData);
#else
        createBuffer(1, VK_BUFFER_USAGE_VERTEX_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_testBuffer, m_testBufferMemory);
        vkMapMemory(m_device, m_testBufferMemory, 0, 1, 0, &m_testBufferData);
#endif
    }
};

int main(int argc, char* argv[])
{
    Application app;
    
    app.init();
    app.mapPersistentMemory();
    app.createTextureImage();
    app.mainLoop();

    return 0;
}
